#include "DCMotor.h"

DCMotor::DCMotor(PCA9685* pwmCtrlr, uint8_t ch1, uint8_t ch2, uint8_t minSpeed, bool normal_direction, uint8_t inert_factor) {
	this->pwmController = pwmCtrlr;
	this->channelOne = ch1;
	this->channelTwo = ch2;
	this->minSpeed = minSpeed;
	this->normalDirection = normal_direction;
	this->inertFactor = inert_factor;
}

void DCMotor::attachEncoder(OpticalEncoder* opticalEncoder) {
	this->opticalEncoder = opticalEncoder;
}

uint8_t DCMotor::getSpeed() {
	if (this->forwardSpeed > 0) {
		return this->forwardSpeed;
	} else if(this->backwardSpeed > 0) {
		return this->backwardSpeed;
	}

	return 0;
}

bool DCMotor::isStopped() {
	return this->state == STATE_STOPPED;
}

bool DCMotor::isForward() {
	return this->state == STATE_FORWARD;
}

bool DCMotor::isBackward() {
	return this->state == STATE_BACKWARD;
}

bool DCMotor::isBusy() {
	#ifdef DEBUG_MODE
		Serial.print("\nDCMotor (");Serial.print(this->id);Serial.print(")->isBusy(): ");
	#endif

	if (
			(	this->opticalEncoder
				&& this->opticalEncoder->isLimited()
				&& this->opticalEncoder->isLimitExceeded() == false
			)
			|| this->task != MOTOR_TASK_IDLE

	) {
		#ifdef DEBUG_MODE
			Serial.println("I'm busy");
		#endif

		return true;
	} else {
		#ifdef DEBUG_MODE
			Serial.println("I'm free");
		#endif

		return false;
	}
}

void DCMotor::setForwardSpeed(uint8_t speed) {
	#ifdef DEBUG_MODE
		Serial.print("\nDCMotor (");Serial.print(this->id);Serial.print(")->setForwardSpeed(): ");
		Serial.println(speed);
	#endif

	speed = this->normalizeSpeed(speed);
	this->state = STATE_FORWARD;

	this->backwardSpeed = 0;
	
	if (this->forwardSpeed == speed) {
		return;
	}

	this->forwardSpeed = speed;

	if (this->normalDirection) {
		this->pwmController->setRelPWM(this->channelOne, speed);
		this->pwmController->setRelPWM(this->channelTwo, 0);
	} else {
		this->pwmController->setRelPWM(this->channelOne, 0);
		this->pwmController->setRelPWM(this->channelTwo, speed);
	}
}

void DCMotor::forward(uint8_t speed) {
	if (this->isBusy()) {
		return;
	}

	this->encoderEnabled = false;
	this->setForwardSpeed(speed);
}

void DCMotor::forward(uint8_t speed, float rotations) {
	if (this->isBusy()) {
		return;
	}

	if (rotations > 0 && this->opticalEncoder) {
		this->encoderEnabled = true;

		this->opticalEncoder->
			setLimit(
				(uint16_t)this->opticalEncoder->sectorsPerLoop * rotations,
				true
			);

		this->setForwardSpeed(speed);
	}
}

void DCMotor::setBackwardSpeed(uint8_t speed) {
	#ifdef DEBUG_MODE
		Serial.print("\nDCMotor (");Serial.print(this->id);Serial.print(")->setBackwardSpeed(): ");
		Serial.println(speed);
	#endif

	speed = this->normalizeSpeed(speed);
	this->state = STATE_BACKWARD;

	this->forwardSpeed = 0;

	if (this->backwardSpeed == speed) {
		return;
	}

	this->backwardSpeed = speed;
	
	if (this->normalDirection) {
		this->pwmController->setRelPWM(this->channelOne, 0);
		this->pwmController->setRelPWM(this->channelTwo, speed);
	} else {
		this->pwmController->setRelPWM(this->channelOne, speed);
		this->pwmController->setRelPWM(channelTwo, 0);
	}
}

void DCMotor::backward(uint8_t speed) {
	if (this->isBusy()) {
		return;
	}

	this->encoderEnabled = false;
	this->setBackwardSpeed(speed);
}

void DCMotor::backward(uint8_t speed, float rotations) {
	if (this->isBusy()) {
		return;
	}

	if (rotations > 0 && opticalEncoder) {
		this->encoderEnabled = true;
		this->opticalEncoder->
			setLimit(
				(uint16_t)this->opticalEncoder->sectorsPerLoop * rotations,
				true
			);

		this->setBackwardSpeed(speed);
	}
}

void DCMotor::stop(bool hard) {
	#ifdef DEBUG_MODE
		Serial.print("\nDCMotor (");Serial.print(this->id);Serial.print(")->stop(): ");
		Serial.println(speed);
	#endif

	if (hard && (forwardSpeed > 0 || backwardSpeed > 0)) {
		#ifdef DEBUG_MODE
			Serial.println("HARD");
		#endif

		uint16_t stopTime = getHardStopTime();

		if (forwardSpeed > 0) {
			this->state = STATE_FORWARD;
			this->backward(100);
			this->dalayAndOff(stopTime);

		} else if (backwardSpeed > 0) {
			this->state = STATE_BACKWARD;
			this->forward(100);
			this->dalayAndOff(stopTime);
		}		
	} else {
		#ifdef DEBUG_MODE
			Serial.println("SOFT");
		#endif

		this->off();
	}
}

// @periodMs - 1 tick = 1ms. So max tick size can be 255 * 1ms = 255ms
void DCMotor::externalTick(uint8_t &periodMs) {
	if (this->task != MOTOR_TASK_IDLE) {
		tickCounter += periodMs;

		if (tickCounter >= tickLimit) {
			tickCounter = 0;
			tickLimit = 0;
			this->task = MOTOR_TASK_IDLE;

			if (this->task == MOTOR_TASK_OFF) {
				off();
			}
		}
	}

	if (this->opticalEncoder
		&& this->state != STATE_STOPPED
	) {
		this->checkEncoderLimit();
	}
}

void DCMotor::abort() {
	if (this->opticalEncoder) {
		this->opticalEncoder->disable();
		this->opticalEncoder->reset();
	}

	//	this->encoderEnabled = false;
	this->state = STATE_STOPPED;
	this->task = MOTOR_TASK_IDLE;
}

/*****************************************************************/
/***********************     Private     *************************/
/*****************************************************************/

void DCMotor::checkEncoderLimit() {
	if (isEncoderLimitExceeded()) {
		this->stop(false);

		#ifdef DEBUG_MODE
			Serial.print("\nDCMotor (");Serial.print(this->id);Serial.print(")->checkEncoderLimit(): ");
			Serial.println("Disable encoder of motor");
		#endif

		this->opticalEncoder->disable();
		this->opticalEncoder->reset();
	}
}

bool DCMotor::isEncoderLimitExceeded() {
	return this->encoderEnabled
		&& this->opticalEncoder
		&& (this->opticalEncoder->isLimitExceeded());
}

void DCMotor::off() {
	this->state = STATE_STOPPED;
	this->forwardSpeed = 0;
	this->backwardSpeed = 0;
	this->pwmController->setRelPWM(this->channelOne, 0);
	this->pwmController->setRelPWM(this->channelTwo, 0);

	if (opticalEncoder) {
		#ifdef DEBUG_MODE
			Serial.print("\nDCMotor (");Serial.print(this->id);Serial.print(")->off(): ");
			Serial.println("Disable encoder of motor");
		#endif

		opticalEncoder->disable();
	}
}

void DCMotor::dalayAndOff(uint8_t time_ms) {
	this->tickLimit = time_ms;
	this->task = MOTOR_TASK_OFF;
}


uint8_t DCMotor::getHardStopTime() {
	uint8_t speed = max(this->backwardSpeed, this->forwardSpeed);
	return (uint8_t)(( (speed/2 + (100-speed)) * inertFactor)/100);
}

uint8_t DCMotor::normalizeSpeed(uint8_t speed) {
	speed = constrain(speed, 0, 100);
	return speed == 0 ? 0 : map(speed, 0, 100, this->minSpeed, 100);
}
