#include "DualWheelSpeedDifferenceRegulator.h"

DualWheelSpeedDifferenceRegulator::DualWheelSpeedDifferenceRegulator() {
	this->pidRegulator = new PidRegulator();
	this->pidRegulator->setResultLimits(MIN_CORRECTION_LIMIT, MAX_CORRECTION_LIMIT);
}

DualWheelSpeedDifferenceRegulator::DualWheelSpeedDifferenceRegulator(DCMotor* leftMotor, DCMotor* rightMotor, OpticalEncoder* leftEncoder, OpticalEncoder* rightEncoder) {
	this->pidRegulator = new PidRegulator();
	this->pidRegulator->setResultLimits(MIN_CORRECTION_LIMIT, MAX_CORRECTION_LIMIT);

	this->leftMotor = leftMotor;
	this->rightMotor = rightMotor;
	this->leftOpticalEncoder = leftEncoder;
	this->rightOpticalEncoder = rightEncoder;
}

void DualWheelSpeedDifferenceRegulator::setDebugMode(char mode) {
	this->pidRegulator->setDebugMode(mode);
}

void DualWheelSpeedDifferenceRegulator::attachLeftMotor(DCMotor* leftMotor) {
	this->leftMotor = leftMotor;
}

void DualWheelSpeedDifferenceRegulator::attachRightMotor(DCMotor* rightMotor) {
	this->rightMotor = rightMotor;
}

void DualWheelSpeedDifferenceRegulator::attachLeftOpticalEncoder(OpticalEncoder* leftEncoder) {
	this->leftOpticalEncoder = leftEncoder;
}

void DualWheelSpeedDifferenceRegulator::attachRightOpticalEncoder(OpticalEncoder* rightEncoder) {
	this->rightOpticalEncoder = rightEncoder;
}

void DualWheelSpeedDifferenceRegulator::reset() {
	this->pidRegulator->reset();
}

void DualWheelSpeedDifferenceRegulator::setSpeedDifference(int8_t leftSpeed, int8_t rightSpeed) {
	this->leftSpeed = leftSpeed;
	this->rightSpeed = rightSpeed;
	
	if (leftSpeed < 0 && rightSpeed > 0) {
		this->pidRegulator->setResultLimits(-(abs(leftSpeed) + rightSpeed), abs(leftSpeed) + rightSpeed);
	} else if (leftSpeed > 0 && rightSpeed > 0) {
		this->pidRegulator->setResultLimits(abs(leftSpeed - rightSpeed) - 100, 100 - abs(leftSpeed - rightSpeed));
	} else if (leftSpeed > 0 && rightSpeed < 0) {
		this->pidRegulator->setResultLimits(-(leftSpeed + abs(rightSpeed)), leftSpeed + abs(rightSpeed));
	}
}

void DualWheelSpeedDifferenceRegulator::setPidFactors(float Kp, float Ki, float Kd) {
	this->pidRegulator->setPidFactors(Kp, Ki, Kd);
}

void DualWheelSpeedDifferenceRegulator::run() {
	#ifdef DEBUG_MODE
		Serial.println("\n2WD_SPEED_DIFF_REGULTOR [");
	#endif

	// Check if one or both encoders are exceeded, than return
	// because too late and we can't to regulate anything
	if (leftOpticalEncoder->isLimited() == false
		|| leftOpticalEncoder->isLimitExceeded()
		|| rightOpticalEncoder->isLimited() == false
		|| rightOpticalEncoder->isLimitExceeded()
	) {
		#ifdef DEBUG_MODE
			Serial.print("exit 1\n]");
		#endif

		return;
	}
	
	// Check if still no data from encoders,
	// than need return and try again later
	if (leftOpticalEncoder->getPulses() == 0
		&& rightOpticalEncoder->getPulses() == 0
	) {
		#ifdef DEBUG_MODE
			Serial.print("exit 2\n]");
		#endif

		return;
	}

	#ifdef DEBUG_MODE
		Serial.print("Initial speed: L="); Serial.print(this->leftSpeed);
		Serial.print(", R="); Serial.print(this->rightSpeed);
	#endif

	int16_t lSpeed = 0;
	int16_t rSpeed = 0;
	uint16_t speedDifference = 0;

	if (this->rightSpeed > 0) {
		speedDifference = abs((this->leftSpeed * 100)/this->rightSpeed);
	}

	uint16_t currentDifference = (abs(leftOpticalEncoder->getPulses()) * 100)/max(1, abs(rightOpticalEncoder->getPulses()));

	int16_t correction = this->pidRegulator->run((uint16_t)currentDifference, speedDifference);

	#ifdef DEBUG_MODE
		Serial.print("\tDiff: Want="); Serial.print(speedDifference);
		Serial.print(", Real="); Serial.print(currentDifference);
		Serial.print("\n\tPulses: L="); Serial.print(this->leftOpticalEncoder->getPulses());
		Serial.print(", R="); Serial.print(this->rightOpticalEncoder->getPulses());
		Serial.print(", Corr="); Serial.print(correction);
		Serial.print(correction > 0 ? " L>R" : " L<R");
	#endif

	if (correction > 0) {
		// Left faster
		int16_t leftRange = abs(this->leftSpeed);
		int16_t rightRange = 100 - abs(this->rightSpeed);

		if (this->leftSpeed > 0) {
			lSpeed = this->leftSpeed - (leftRange * correction/(leftRange + rightRange));
		} else {
			lSpeed = this->leftSpeed + (leftRange * correction/(leftRange + rightRange));
		}

		if (this->rightSpeed > 0) {
			rSpeed = this->rightSpeed + (rightRange * correction/(leftRange + rightRange));
		} else {
			rSpeed = this->rightSpeed - (rightRange * correction/(leftRange + rightRange));
		}
		
	} else {
		// Right faster
		int16_t leftRange = 100 - abs(this->leftSpeed);
		int16_t rightRange = abs(this->rightSpeed);

		if (this->leftSpeed > 0) {
			lSpeed = this->leftSpeed + abs((leftRange * correction/(leftRange + rightRange)));
		} else {
			lSpeed = this->leftSpeed - abs((leftRange * correction/(leftRange + rightRange)));
		}

		if (this->rightSpeed > 0) {
			rSpeed = this->rightSpeed - abs((rightRange * correction/(leftRange + rightRange)));
		} else {
			rSpeed = this->rightSpeed + abs((rightRange * correction/(leftRange + rightRange)));
		}
	}

	if (this->leftSpeed > 0) {
		lSpeed = constrain(lSpeed, 0, 100);
	} else {
		lSpeed = constrain(lSpeed, -100, 0);
	}
	
	if (this->rightSpeed > 0) {
		rSpeed = constrain(rSpeed, 0, 100);
	} else {
		rSpeed = constrain(rSpeed, -100, 0);
	}

	this->setMotorSpeed(leftMotor, lSpeed);
	this->setMotorSpeed(rightMotor, rSpeed);

	#ifdef DEBUG_MODE
		Serial.print("\n\tCALCULATE NEW SPEED: newL=");  Serial.print(lSpeed);
		Serial.print(", newR="); Serial.print(rSpeed); Serial.print("\n]");
	#endif
}

void DualWheelSpeedDifferenceRegulator::setMotorSpeed(DCMotor* motor, int8_t speed) {
	if (speed == 0) {
		if (motor->isForward()) {
			motor->setForwardSpeed(0);
		} else {
			motor->setBackwardSpeed(0);
		}
	} else if (speed > 0) {
		if (motor->isForward()) {
			motor->setForwardSpeed(speed);
		} else {
			motor->setForwardSpeed(0);
		}
	} else {

		if (motor->isBackward()) {
			motor->setBackwardSpeed(abs(speed));
		} else {
			motor->setBackwardSpeed(0);
		}
	}
}


