#include "DualWheelSpeedRegulator.h"

DualWheelSpeedRegulator::DualWheelSpeedRegulator() {
	this->pidRegulator = new PidRegulator();
	this->pidRegulator->setResultLimits(MIN_CORRECTION_LIMIT, MAX_CORRECTION_LIMIT);
}

DualWheelSpeedRegulator::DualWheelSpeedRegulator(DCMotor* leftMotor, DCMotor* rightMotor, OpticalEncoder* leftEncoder, OpticalEncoder* rightEncoder) {
	this->pidRegulator = new PidRegulator();
	this->pidRegulator->setResultLimits(MIN_CORRECTION_LIMIT, MAX_CORRECTION_LIMIT);

	this->leftMotor = leftMotor;
	this->rightMotor = rightMotor;
	this->leftOpticalEncoder = leftEncoder;
	this->rightOpticalEncoder = rightEncoder;
}

void DualWheelSpeedRegulator::setDebugMode(char mode) {
	this->pidRegulator->setDebugMode(mode);
}

void DualWheelSpeedRegulator::attachLeftMotor(DCMotor* leftMotor) {
	this->leftMotor = leftMotor;
}

void DualWheelSpeedRegulator::attachRightMotor(DCMotor* rightMotor) {
	this->rightMotor = rightMotor;
}

void DualWheelSpeedRegulator::attachLeftOpticalEncoder(OpticalEncoder* leftEncoder) {
	this->leftOpticalEncoder = leftEncoder;
}

void DualWheelSpeedRegulator::attachRightOpticalEncoder(OpticalEncoder* rightEncoder) {
	this->rightOpticalEncoder = rightEncoder;
}

void DualWheelSpeedRegulator::reset() {
	this->pidRegulator->reset();
}

void DualWheelSpeedRegulator::setSpeed(int8_t speed) {
	this->speed = speed;
}

void DualWheelSpeedRegulator::setPidFactors(float Kp, float Ki, float Kd) {
	this->pidRegulator->setPidFactors(Kp, Ki, Kd);
}

void DualWheelSpeedRegulator::run() {
	#ifdef DEBUG_MODE
		Serial.println("\n2WD_SPEED_REGULTOR [");
	#endif

	// Check if one or both encoders are exceeded, than return
	// because too late and we can't to regulate anything
	if (this->leftOpticalEncoder->isLimited() == false
		|| this->leftOpticalEncoder->isLimitExceeded()
		|| this->rightOpticalEncoder->isLimited() == false
		|| this->rightOpticalEncoder->isLimitExceeded()
	) {
		#ifdef DEBUG_MODE
			Serial.print("exit 1\n]");
		#endif

		return;
	}

	// Check if still no data from encoders,
	// than need return and try again later
	if (this->leftOpticalEncoder->getPulses() == 0
		&& this->rightOpticalEncoder->getPulses() == 0
	) {
		#ifdef DEBUG_MODE
			Serial.print("exit 2\n]");
		#endif

		return;
	}

	uint16_t leftSpeed = 0;
	uint16_t rightSpeed = 0;
	
	int16_t correction = this->pidRegulator->run(
			(int16_t)leftOpticalEncoder->getPulses(),
			(int16_t)rightOpticalEncoder->getPulses()
		);

	if (correction > 0) {
		// Left faster
		int16_t leftRange = speed;
		int16_t rightRange = 100 - speed;

		leftSpeed = speed - (leftRange * correction/(leftRange + rightRange));
		rightSpeed = speed + (rightRange * correction/(leftRange + rightRange));
	} else {
		// Right faster
		uint16_t leftRange = 100 - speed;
		uint16_t rightRange = speed;

		leftSpeed = speed + (leftRange * abs(correction)/(leftRange + rightRange));
		rightSpeed = speed - (rightRange * abs(correction)/(leftRange + rightRange));
	}

	leftSpeed = constrain(leftSpeed, 0, 100);
	rightSpeed = constrain(rightSpeed, 0, 100);

	#ifdef DEBUG_MODE
		Serial.print("\tCorr:"); Serial.print(correction);
		Serial.print(", New left="); Serial.print(leftSpeed);
		Serial.print("right="); Serial.print(rightSpeed); Serial.println("\n]");
	#endif

	if (speed < 0) {
		if (leftMotor->isStopped() == false) {
			leftMotor->setBackwardSpeed(leftSpeed);
		}
		
		if (rightMotor->isStopped() == false) {
			rightMotor->setBackwardSpeed(rightSpeed);
		}
	} else {
		if (leftMotor->isStopped() == false) {
			leftMotor->setForwardSpeed(leftSpeed);
		}

		if (rightMotor->isStopped() == false) {
			rightMotor->setForwardSpeed(rightSpeed);
		}
	}
}
