#include "DualWheelTruck.h"

DualWheelTruck* DualWheelTruck::instance = nullptr;

DualWheelTruck::DualWheelTruck(uint8_t PCA9685_address, uint8_t wheelRadius_mm, uint16_t wheelBase_mm, uint8_t pwmFreq) {
	this->wheelRadius = wheelRadius_mm;
	this->wheelBase = wheelBase_mm;
	this->pwmController = new PCA9685(PCA9685_address);
	this->pwmFrequency = pwmFreq;
	DualWheelTruck::instance = this;
}

void DualWheelTruck::asyncRun() {
	if (this->leftMotor) {
		this->leftMotor->id = 1;
		this->leftMotor->externalTick(this->timerDelayMS);
	}

	if (this->rightMotor) {
		this->rightMotor->externalTick(this->timerDelayMS);
	}

	this->speedRegulatorCallTimer++;

	if (this->speedRegulatorCallTimer >= this->speedRegulatorCallPeriod) {
		this->speedRegulatorCallTimer = 0;
		this->correctSpeed();
	}
}

void DualWheelTruck::init() {
	this->pwmController->begin();

	// Set PWM ferquency. 
	// Attention! If you use servos, than PWMFreq must be set as 50HZ!
	this->pwmController->setPWMFreq(this->pwmFrequency);

	if (!this->speedRegulator) {
		this->speedRegulator = new DualWheelSpeedRegulator();
	}

	if (!this->speedDiffRegulator) {
		this->speedDiffRegulator = new DualWheelSpeedDifferenceRegulator();
	}

	// Init interrupts for motors's and PID regulator's and etc.
	Chronos::attachMsCallback(timerHandler, (uint16_t)this->timerDelayMS);
}

void DualWheelTruck::timerHandler(void) {
	DualWheelTruck::me()->asyncRun();
}

void DualWheelTruck::setSpeedAlignmentPidFactors(float Kp, float Ki, float Kd) {
	if (this->speedRegulator) {
		this->speedRegulator->setPidFactors(Kp, Ki, Kd);
	}
}

void DualWheelTruck::setSpeedDifferencePidFactors(float Kp, float Ki, float Kd) {
	if (this->speedDiffRegulator) {
		this->speedDiffRegulator->setPidFactors(Kp, Ki, Kd);
	}
}

void DualWheelTruck::setSpeedAlignmentPidDebugMode(char mode) {
	if (this->speedRegulator) {
		this->speedRegulator->setDebugMode(mode);
	}
}

void DualWheelTruck::setSpeedDifferencePidDebugMode(char mode) {
	if (this->speedDiffRegulator) {
		this->speedDiffRegulator->setDebugMode(mode);
	}
}

void DualWheelTruck::attachLeftMotor(uint8_t ch1, uint8_t ch2, uint8_t minSpeed, uint8_t inertFactor, bool normalDirection) {
	this->leftMotor = new DCMotor(this->pwmController, ch1, ch2, minSpeed, normalDirection, inertFactor);
	this->leftMotor->id = 1;

	if (this->speedRegulator) {
		this->speedRegulator->attachLeftMotor(leftMotor);
	}

	if (this->speedDiffRegulator) {
		this->speedDiffRegulator->attachLeftMotor(leftMotor);
	}
}

void DualWheelTruck::attachRightMotor(uint8_t ch1, uint8_t ch2, uint8_t minSpeed, uint8_t inertFactor, bool normalDirection) {
	this->rightMotor = new DCMotor(this->pwmController, ch1, ch2, minSpeed, normalDirection, inertFactor);
	this->rightMotor->id = 2;

	if (this->speedRegulator) {
		this->speedRegulator->attachRightMotor(rightMotor);
	}

	if (this->speedDiffRegulator) {
		this->speedDiffRegulator->attachRightMotor(rightMotor);
	}
}

void DualWheelTruck::attachLeftOpticalEncoder(uint8_t ch, uint8_t sectorsPerLoop) {
	this->leftOpticalEncoder = new OpticalEncoder(ch, sectorsPerLoop, TWO_PI*this->wheelRadius);
	this->leftMotor->attachEncoder(this->leftOpticalEncoder);

	if (this->speedRegulator) {
		this->speedRegulator->attachLeftOpticalEncoder(leftOpticalEncoder);
	}

	if (this->speedDiffRegulator) {
		this->speedDiffRegulator->attachLeftOpticalEncoder(leftOpticalEncoder);
	}
}

void DualWheelTruck::attachRightOpticalEncoder(uint8_t ch, uint8_t sectorsPerLoop) {
	this->rightOpticalEncoder = new OpticalEncoder(ch, sectorsPerLoop, TWO_PI*this->wheelRadius);
	this->rightMotor->attachEncoder(this->rightOpticalEncoder);

	if (this->speedRegulator) {
		this->speedRegulator->attachRightOpticalEncoder(this->rightOpticalEncoder);
	}

	if (this->speedDiffRegulator) {
		this->speedDiffRegulator->attachRightOpticalEncoder(this->rightOpticalEncoder);
	}
}

// ****************************************************************
// ************************  LEFT MOTOR   *************************
// ****************************************************************

void DualWheelTruck::leftMotorForward(uint8_t speed, uint16_t distance) {
	if (this->leftMotor) {
		if (distance > 0) {
			#ifdef DEBUG_MODE
				Serial.print("\nLeftMotorForward [\ndest="); Serial.print(distance);
				Serial.print("\nsectors=");
				Serial.print(
					(uint16_t)this->leftOpticalEncoder->sectorsPerLoop
						* this->distanceToWheelRotations(distance)
				);
				Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
				Serial.println("]");
			#endif

			this->leftMotor->forward(speed, this->distanceToWheelRotations(distance));
		} else {
			this->leftMotor->forward(speed);
		}

		this->disableSpeedCorrection();
	}
}

void DualWheelTruck::leftMotorBackward(uint8_t speed, uint16_t distance) {
	if (this->leftMotor) {
		if (distance > 0) {
			#ifdef DEBUG_MODE
				Serial.print("\nLeftMotorBackward [\ndest="); Serial.print(distance);
				Serial.print("\nsectors=");
				Serial.print(
					(uint16_t)this->leftOpticalEncoder->sectorsPerLoop
						* this->distanceToWheelRotations(distance)
				);
				Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
				Serial.println("]");
			#endif

			this->leftMotor->backward(speed, this->distanceToWheelRotations(distance));
		} else {
			this->leftMotor->backward(speed);
		}

		this->disableSpeedCorrection();
	}
}

void DualWheelTruck::leftMotorStop(bool hard) {
	if (this->leftMotor) {
		this->leftMotor->stop(hard);
		this->disableSpeedCorrection();
	}
}

// ****************************************************************
// ***********************  RIGHT MOTOR   *************************
// ****************************************************************

void DualWheelTruck::rightMotorForward(uint8_t speed, uint16_t distance) {
	if (this->rightMotor) {
		if (distance > 0) {
			#ifdef DEBUG_MODE
				Serial.print("\nRghtMotorForward [\ndest="); Serial.print(distance);
				Serial.print("\nsectors=");
				Serial.print(
					(uint16_t)this->rightOpticalEncoder->sectorsPerLoop
						* this->distanceToWheelRotations(distance)
				);
				Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
				Serial.println("]");
			#endif

			this->rightMotor->forward(speed, this->distanceToWheelRotations(distance));
		} else {
			this->rightMotor->forward(speed);
		}

		this->disableSpeedCorrection();
	}
}

void DualWheelTruck::rightMotorBackward(uint8_t speed, uint16_t distance) {
	if (this->rightMotor) {
		if (distance > 0) {
			#ifdef DEBUG_MODE
				Serial.print("\nRightMotorBackward [\ndest="); Serial.print(distance);
				Serial.print("\nsectors=");
				Serial.print(
					(uint16_t)this->rightOpticalEncoder->sectorsPerLoop
						* this->distanceToWheelRotations(distance)
				);
				Serial.print("\nrot="); Serial.println(this->distanceToWheelRotations(distance));
				Serial.println("]");
			#endif

			this->rightMotor->backward(speed, this->distanceToWheelRotations(distance));
		} else {
			this->rightMotor->backward(speed);
		}

		this->disableSpeedCorrection();
	}
}

void DualWheelTruck::rightMotorStop(bool hard) {
	if (this->rightMotor) {
		this->rightMotor->stop(hard);
		this->disableSpeedCorrection();
	}
}

// ****************************************************************
// **********************  COMMON STUFF   *************************
// ****************************************************************

bool DualWheelTruck::isBusy() {
	return this->leftMotor->isBusy() || this->rightMotor->isBusy();
}

void DualWheelTruck::moveForward(uint8_t speed, uint16_t distance) {
	this->state = MOVE_FORWARD;
	this->leftMotorForward(speed, distance);
	this->rightMotorForward(speed, distance);
	this->enableSpeedCorrection((int8_t)speed);
}

void DualWheelTruck::moveBackward(uint8_t speed, uint16_t distance) {
	this->state = MOVE_BACKWARD;
	this->leftMotorBackward(speed, distance);
	this->rightMotorBackward(speed, distance);
	this->enableSpeedCorrection(-1 * ((int8_t)speed) );
}

void DualWheelTruck::turnLeft(uint8_t speed, uint8_t radius, uint16_t angle) {
	this->state = TURN_LEFT;

	int8_t leftSpeed = 0;

	if (this->leftOpticalEncoder && this->rightOpticalEncoder) {
		// whillBase is in mm, so half of base in cm is: (this->wheelBase/10)/2 => this->wheelBase/20
		uint8_t halfBase = this->wheelBase/20;

		if (radius < halfBase) {
			// Left motor backward
			// Right motor forward
			leftSpeed = (int8_t)((-speed * (halfBase - radius))/(halfBase + radius));

			#ifdef DEBUG_MODE
				Serial.print("\nLeft turn(INNER): [\nleft wheel speed=");
				Serial.print(leftSpeed);
				Serial.print("\n left wheel distance=");
				Serial.print(this->turningAngleToDistance(halfBase - radius, angle));
			#endif

			this->leftMotorBackward(
				leftSpeed, 
				this->turningAngleToDistance(halfBase - radius, angle)
			);
		} else if(radius == halfBase) {
			// Left motor stopped
			// Right motor forward
			#ifdef DEBUG_MODE
				Serial.print("\nLeft turn(ZERO): [\nleft wheel speed=0");
				Serial.print("\nleft wheel distance=0");
			#endif

			this->leftMotorStop(false);
		} else {
			// Left motor forward
			// Right motor forward
			leftSpeed = (int8_t)((speed * (radius - halfBase))/(radius + halfBase));

			#ifdef DEBUG_MODE
				Serial.print("\nLeft turn(OUTER): [\nleft wheel speed=0");
				Serial.print(leftSpeed);
				Serial.print("\nleft wheel distance=0");
				Serial.print(this->turningAngleToDistance(radius - halfBase, angle));
			#endif

			this->leftMotorForward(
				leftSpeed, 
				this->turningAngleToDistance(radius - halfBase, angle)
			);
		}

		#ifdef DEBUG_MODE
			Serial.print("\nright wheel speed=");
			Serial.print(speed);
			Serial.print("\nright wheel distance=0");
			Serial.println(this->turningAngleToDistance(radius + halfBase, angle));
			Serial.println("]");
		#endif

		this->rightMotorForward(
			speed, 
			this->turningAngleToDistance(radius + halfBase, angle)
		);

		if (leftSpeed != 0) {
			this->enableSpeedDifferenceCorrection(leftSpeed, speed);
		}
	}
}

void DualWheelTruck::turnRight(uint8_t speed, uint8_t radius, uint16_t angle) {
	this->state = TURN_RIGHT;

	int8_t rightSpeed = 0;

	if (this->leftOpticalEncoder && this->rightOpticalEncoder) {
		// whillBase is in mm, so half of base in cm is: (this->wheelBase/10)/2 => this->wheelBase/20
		uint8_t halfBase = this->wheelBase/20;

		if (radius < halfBase) {
			// Left motor forward
			// Right motor backward
			rightSpeed = (int8_t)((-speed * (halfBase - radius))/(halfBase + radius));

			#ifdef DEBUG_MODE
				Serial.print("\nRight turn(INNER): [\nright wheel speed=");
				Serial.print(rightSpeed);
				Serial.print("\n right wheel distance=");
				Serial.print(this->turningAngleToDistance(halfBase - radius, angle));
			#endif

			this->rightMotorBackward(
				rightSpeed,
				this->turningAngleToDistance(halfBase - radius, angle)
			);
		} else if(radius == halfBase) {
			// Left motor forward
			// Right motor stopped
			#ifdef DEBUG_MODE
				Serial.print("\nRight turn(ZERO): [\nright wheel speed=0");
				Serial.print("\nright wheel distance=0");
			#endif

			this->rightMotorStop(false);
		} else {
			// Left motor forward
			// Right motor forward
			rightSpeed = (int8_t)((speed * (radius - halfBase))/(radius + halfBase));

			#ifdef DEBUG_MODE
				Serial.print("\nRight turn(OUTER): [\nright wheel speed=0");
				Serial.print(rightSpeed);
				Serial.print("\nright wheel distance=0");
				Serial.print(this->turningAngleToDistance(radius - halfBase, angle));
			#endif

			this->rightMotorForward(
				rightSpeed,
				this->turningAngleToDistance(radius - halfBase, angle)
			);
		}

		#ifdef DEBUG_MODE
			Serial.print("\nleft wheel speed=");
			Serial.print(speed);
			Serial.print("\nleft wheel distance=0");
			Serial.println(this->turningAngleToDistance(radius + halfBase, angle));
			Serial.println("]");
		#endif

		this->leftMotorForward(
			speed, 
			this->turningAngleToDistance(radius + halfBase, angle)
		);

		if (rightSpeed != 0) {
			this->enableSpeedDifferenceCorrection(speed, rightSpeed);
		}
	}
}

// ****************************************************************
// **********************    private     **************************
// ****************************************************************

float DualWheelTruck::distanceToWheelRotations(uint16_t distance) {
	return (float)(distance*10)/(TWO_PI * this->wheelRadius);
}

// Calculate how much wheel rotations need for wheel turn path
float DualWheelTruck::turningAngleToWheelRotations(uint8_t radius, uint16_t angle) {
	// Wheel rotations Wr = L/C, where:
	// L - is wheel path length: L =  2 * Pi * turnRadius * angle/360
	// C - is wheel circle length: C = 2 * Pi * wheelRadius
	// Don't forget that whillRadius is mm scale, so we need devide it by 10.
	// So ttoal equation: Wr = (2 * Pi * turnRadius * angle/360) / (2 * Pi * wheelRadius/10)
	// But if we will reduce 2*Pi from equations 2 and 3, then result 
	// equation will be much simpler: Wr = (turnRadius * angle/360) / (wheelRadius/10) => (turnRadius * angle/36) / wheelRadius
	return (float)(radius * angle/36) / this->wheelRadius;
}

// Calculate how much wheel rotations need for wheel turn path
float DualWheelTruck::turningAngleToDistance(uint8_t radius, uint16_t angle) {
	return (float)(TWO_PI * radius * angle/360);
}

void DualWheelTruck::disableSpeedCorrection() {
	this->speed = 0;
}

void DualWheelTruck::enableSpeedCorrection(int8_t speed) {
	if (this->leftOpticalEncoder
		&& this->rightOpticalEncoder
		&& speed != 0
	) {

		if (this->speedDiffRegulator) {
			this->speedDiffRegulator->reset();
		}

		if (this->speedRegulator) {
			this->speedRegulator->reset();
			this->speedRegulator->setSpeed(speed);
		}

		this->speed = speed;
	}
}

void DualWheelTruck::enableSpeedDifferenceCorrection(int8_t leftSpeed, int8_t rightSpeed) {
	if (this->leftOpticalEncoder
		&& this->rightOpticalEncoder
		&& leftSpeed != 0
		&& rightSpeed != 0
	) {

		if (this->speedRegulator) {
			this->speedRegulator->reset();
		}

		if (this->speedDiffRegulator) {
			this->speedDiffRegulator->reset();
			this->speedDiffRegulator->setSpeedDifference(leftSpeed, rightSpeed);
		}
	}
}

bool DualWheelTruck::isSpeedCorrectionEnabled() {
	return this->speed != 0
		|| this->state == TURN_LEFT
		|| this->state == TURN_RIGHT;
}

void DualWheelTruck::correctSpeed() {
	if (this->isSpeedCorrectionEnabled() == false)
		return;

	if (this->state == MOVE_FORWARD || this->state == MOVE_BACKWARD) {
		if (this->speedRegulator) {
			this->speedRegulator->run();
		}
	}

	if (this->state == TURN_LEFT || this->state == TURN_RIGHT) {
		if (this->speedDiffRegulator) {
			this->speedDiffRegulator->run();
		}
	}
}
