#ifndef _EVERY_2WD_TRUCK_H
#define _EVERY_2WD_TRUCK_H
#include <Arduino.h>
#include <TimerOne.h>
#include "Chronos.h"
#include "PCA9685.h"
#include "DCMotor.h"
#include "OpticalEncoder.h"
#include "PidRegulator.h"
#include "DualWheelSpeedRegulator.h"
#include "DualWheelSpeedDifferenceRegulator.h"
#include "I2CServo.h"
#include <Wire.h>

//#define DEBUG_MODE

enum TruckState {
	STOP,
	MOVE_FORWARD,
	MOVE_BACKWARD,
	TURN_LEFT,
	TURN_RIGHT,
};

class DualWheelTruck {
	public:
		// PWM controller PCA9685
		PCA9685* pwmController;

		DualWheelTruck(uint8_t PCA9685_address, uint8_t wheelRadius_mm, uint16_t wheelBase_mm, uint8_t pwmFreq = 50);

		static DualWheelTruck* instance;

		static DualWheelTruck * me() {
			return instance;
		}

		void init();

		void setSpeedAlignmentPidFactors(float Kp, float Ki, float Kd);
		void setSpeedDifferencePidFactors(float Kp, float Ki, float Kd);
		void setSpeedAlignmentPidDebugMode(char mode);
		void setSpeedDifferencePidDebugMode(char mode);

		void attachLeftMotor(uint8_t ch1, uint8_t ch2, uint8_t minSpeed, uint8_t inertFactor = 0, bool normalDirection = true);
		void attachRightMotor(uint8_t ch1, uint8_t ch2, uint8_t minSpeed, uint8_t inertFactor = 0, bool normalDirection = false);

		void attachLeftOpticalEncoder(uint8_t ch, uint8_t sectorsPerLoop);
		void attachRightOpticalEncoder(uint8_t ch, uint8_t sectorsPerLoop);

		void leftMotorForward(uint8_t speed, uint16_t distance = 0);
		void leftMotorBackward(uint8_t speed, uint16_t distance = 0);
		void leftMotorStop(bool hard = true);

		void rightMotorForward(uint8_t speed, uint16_t distance = 0);
		void rightMotorBackward(uint8_t speed, uint16_t distance = 0);
		void rightMotorStop(bool hard = true);

		void moveForward(uint8_t speed, uint16_t distance = 0);
		void moveBackward(uint8_t speed, uint16_t distance = 0);
		void turnLeft(uint8_t speed, uint8_t radius, uint16_t angle);
		void turnRight(uint8_t speed, uint8_t radius, uint16_t angle);
		bool isBusy();

		static void timerHandler(void);

	private:
		uint8_t pwmFrequency = 50;

		TruckState state = STOP;

		// Motors
		DCMotor* leftMotor;
		DCMotor* rightMotor;

		// Encoders
		OpticalEncoder* leftOpticalEncoder;
		OpticalEncoder* rightOpticalEncoder;

		// 2 WD forward-backward speed regulator
		DualWheelSpeedRegulator* speedRegulator;
		// 2 WD turn left-right speed regulator
		DualWheelSpeedDifferenceRegulator* speedDiffRegulator;

		uint16_t leftEncoderCounter = 0;
		uint16_t rightEncoderCounter = 0;

		// Remember speed. Signed value from -100% to 100%
		int8_t speed = 0;
		uint8_t speedRegulatorCallPeriod = 20;
		uint8_t speedRegulatorCallTimer = 0;
		
		uint8_t timerDelayMS = 1;

		uint8_t wheelRadius = 0; // millimeters
		uint16_t wheelBase = 0; // millimeters

		void asyncRun();
		// void setBusy(bool);
		void correctSpeed();
		void disableSpeedCorrection();
		void enableSpeedCorrection(int8_t speed);
		void enableSpeedDifferenceCorrection(int8_t leftSpeed, int8_t rightSpeed);
		bool isSpeedCorrectionEnabled();
		float distanceToWheelRotations(uint16_t distance);
		float turningAngleToWheelRotations(uint8_t radius, uint16_t angle);
		float turningAngleToDistance(uint8_t radius, uint16_t angle);
};

#endif
