#include "OpticalEncoder.h"

OpticalEncoder::OpticalEncoder(uint8_t channel, uint8_t sectorsPerLoop, uint16_t wheelLength) {
	this->channel = channel;
	this->sectorsPerLoop = sectorsPerLoop;
	this->wheelLength = wheelLength;
	this->sectorAngle = 360/sectorsPerLoop;
	Chronos::attachMsMemberCallback(this, 1);
}

void OpticalEncoder::chronosHandler() {
	this->update();
}

void OpticalEncoder::enable() {
	#ifdef DEBUG_MODE
		Serial.print("\OpticalEncoder (");Serial.print((int)this);Serial.println(")->enable(): OK");
	#endif

	this->enabled = true;
}

void OpticalEncoder::disable() {
	#ifdef DEBUG_MODE
		Serial.print("\OpticalEncoder (");Serial.print((int)this);Serial.println(")->disable(): OK");
	#endif

	this->enabled = false;
}

void OpticalEncoder::setLimit(uint16_t limit, bool isForward) {
	#ifdef DEBUG_MODE
		Serial.print("\OpticalEncoder (");Serial.print((int)this);Serial.print(")->setLimit(): ");
		Serial.println(limit);
	#endif

	this->reset();
	this->isForward = isForward;
	this->rotationLimit = limit;
	this->enable();
}

bool OpticalEncoder::isLimitExceeded() {
	return this->rotationLimit != 0 && abs(counter) >= this->rotationLimit;
}

bool OpticalEncoder::isLimited() {
	return this->rotationLimit > 0;
}

void OpticalEncoder::update() {
	if (this->enabled == false)
		return;

//	Serial.print("*");

	uint8_t curState = digitalRead(this->channel);
	this->timerCounter += TIMER_PERIOD;
	this->samplesCounter++;

	if (curState != this->state) {

		if (this->filterCounter < FILTER_COUNTER_LIMIT) {
			this->filterCounter++;

		} else {
			this->state = curState;
			this->filterCounter = 0;
			this->pulsesCounter++;

			if (this->isForward == true) {
				this->counter++;
			} else {
				this->counter--;
			}
		}
	}

	if (this->samplesCounter == 25 || this->timerCounter > TIMER_LIMIT) {
		this->fillBuffer();
		this->samplesCounter = 0;
	}
}

int32_t& OpticalEncoder::getPulses() {
	return this->counter;
}

void OpticalEncoder::getPulsesFrequency(uint16_t * freq) {
	uint16_t numberOfPulses = 0;

	// Tick's. Not seconds.
	uint16_t timeInterval = 0;

	for (int i = 0; i < this->bufferSize; ++i) {
		numberOfPulses += this->speedBuffer[i];
		timeInterval += this->timeBuffer[i];
	}

	if (timeInterval > 0) {
		*freq = (TICK_FACTOR/timeInterval) * numberOfPulses;
	} else {
		*freq = 0;
	}
}

void OpticalEncoder::getAngleSpeed(uint16_t * speed) {
	uint16_t numberOfSectors = 0;
	uint16_t timeInterval = 0;

	for (int i = 0; i < this->bufferSize; ++i) {
		numberOfSectors += this->speedBuffer[i];
		timeInterval += this->timeBuffer[i];
	}

	*speed = (float)(numberOfSectors * this->sectorAngle) / ((float)timeInterval/1000);
}

// cm/sec
void OpticalEncoder::getSpeed(uint16_t * speed) {
	this->getAngleSpeed(speed);
	*speed = wheelLength*(*speed/360)/10;
}

void OpticalEncoder::fillBuffer() {
	if (this->bufferIndex == this->bufferSize) {
		this->bufferIndex = 0;
	}

	if (this->bufferDataLength < this->bufferSize) {
		this->bufferDataLength++;
	}

	this->speedBuffer[this->bufferIndex] = this->pulsesCounter;
	this->timeBuffer[this->bufferIndex] = this->timerCounter;
	this->bufferIndex++;
	this->pulsesCounter = 0;
	this->timerCounter = 0;
}

void OpticalEncoder::reset() {
	this->state = digitalRead(this->channel);
	this->counter = 0;
	this->timerCounter = 0;
	this->bufferIndex = 0;
	this->bufferDataLength = 0;
	this->samplesCounter = 0;
	this->rotationLimit = 0;

	for (this->pulsesCounter = 0; this->pulsesCounter < this->bufferSize; this->pulsesCounter++) {
		this->speedBuffer[this->pulsesCounter] = 0;
		this->timeBuffer[this->pulsesCounter] = 0;
	}

	this->pulsesCounter = 0;
}
