#include <Wire.h>
#include <I2C.h>
#include <I2CTransport.h>
#include <PCA9685.h>
#include <DualWheelTruck.h>

#define LED_PIN 13
#define LEFT_ENCODER_PIN 9
#define RIGHT_ENCODER_PIN 6

int encoderState = HIGH;

I2C myI2C = I2C();
I2CTransport transport = I2CTransport();

// Создаем объект - 2-х колесная тележка:
// [I2C адрес микросхемы PCA9685]
// [радиус колеса, мм]
// [ширина колесной базы, мм]
// [частота ШИМ для управления двигателями. Если используете сервоприводы, выбирайте 50Гц]
DualWheelTruck truck = DualWheelTruck(0x40, 35, 170, 100);

uint8_t task = 0;

void setup() {
  I2CTransport::setClient(&myI2C);
  I2CTransport::begin();

  pinMode(LEFT_ENCODER_PIN, INPUT);
  pinMode(RIGHT_ENCODER_PIN, INPUT);
  pinMode(LED_PIN, OUTPUT);
  Serial.begin(9600);
  delay(1000);
  
  // Создаем объект truck - это главный оъект, то есть шасси.
  // Серез него общаемся с остальными объектами если надо
  truck.init();

  // Создаем моторы со следующими аргументами:
  // [номер первого порта мотора (у микросхемы PCА9685 а не у контроллера)]
  // [номер второго порта мотора (у микросхемы PCА9685 а не у контроллера)]
  // [инертность колеса. Может принимать значения от 0 до бесконечности, но в реальности больше 10 точно не надо] - на основе 
  // этой величины рассчитывается длительностьтормозного импульса. Величина безразмерная и подбирается экспериментально.
  // [полярность подключения мотора - прямая или реверсивная] - нужна чтобы не перекручивать провода мотора
  truck.attachLeftMotor(0, 1, 3, 25, false);
  truck.attachRightMotor(2, 3, 3, 25, true);

  // Создаем энкодеры для левого и правого моторов с аргументами:
  // [номер порта к которому подключен энкодер]
  // [кол-во секторов на один оборот диска энкодера]
  truck.attachLeftOpticalEncoder(LEFT_ENCODER_PIN, 24);
  truck.attachRightOpticalEncoder(RIGHT_ENCODER_PIN, 24);

  // На всякий случай тормозим оба мотора. 
  // Ведь если они были запущены то микросхема PCA не изменит своего состояния? 
  // если не снять питание. Даже если сделать контроллеру reset.
  truck.leftMotorStop(true);
  truck.rightMotorStop(true);

  // Настройка пид регулятор ответственный за равномерность хода во
  // время прямолинейного движения
  // [Kp - коэф пропорциональной составляющей]
  // [Ki - коэф интегральной составляющей]
  // [Kd - коэф дифференциальной составляющей]
  truck.setSpeedAlignmentPidFactors(30, 0.1, 10);

  // Настройка пид регулятор ответственный за равномерность хода во
  // время совершения поворотов
  // [Kp - коэф пропорциональной составляющей]
  // [Ki - коэф интегральной составляющей]
  // [Kd - коэф дифференциальной составляющей]
  truck.setSpeedDifferencePidFactors(10, 0.1, 5);

  // Настройка логирования PID регулятораю Принимает значение от 0 до 6
  // +----------------------------------------------------------------+
  // | 0 | silent mode                                                |
  // |---+------------------------------------------------------------|
  // | 1 | text output                                                |
  // |---+------------------------------------------------------------|
  // | 2 | print E for visual debug with arduino Serial Plotter       |
  // |---+------------------------------------------------------------|
  // | 3 | print P for visual debug with arduino Serial Plotter       |
  // |---+------------------------------------------------------------|
  // | 4 | print I for visual debug with arduino Serial Plotter       |
  // |---+------------------------------------------------------------|
  // | 5 | print D for visual debug with arduino Serial Plotter       |
  // |---+------------------------------------------------------------|
  // | 6 | print E,P,I,D for visual debug with arduino Serial Plotter |
  // |---+------------------------------------------------------------|
  truck.setSpeedAlignmentPidDebugMode(1);
//  truck.setSpeedDifferencePidDebugMode(1);
  Chronos::delay(3000);
}

void loop() {
  // Проверяем что шасси сейчас не исполняет никакое задание
  if (truck.isBusy()) {
    // Если шасси занато, ждем 100мсек
    Serial.println("I'm busy now");
    digitalWrite(LED_PIN, !digitalRead(LED_PIN));
    Chronos::delay(100);
    return;
  }

  //Если свободно от каких либо задач, продолжаем 
  Serial.println("\nStart new action");
  
  if (task == 0) {
    // Едем вепред на скорости 35%, на расстояние 100см
    Serial.println("= Move forward =");
    /**
     * uint8_t speed
     * uint16_t distance, cm
     */
    truck.moveForward(35, 100);
    task++;
  } else if (task == 1) {
    // Совершаем поворот налево на скорости 35%
    // с радиусом 30см (относительно центра колесной базы)
    // на угол 180 градусов
    Serial.println("= Turn left =");
    /**
     * uint8_t speed
     * uint8_t radius
     * uint16_t angle
     */
    truck.turnLeft(35, 30, 180);
    task = 0;
  }

  /********************************************************************/
  /*********************** Возможные команды **************************/
  /********************************************************************/
  //Управление отдельно левым мотором
  //truck.leftMotorForward(uint8_t speed, uint16_t distance = 0);
  //truck.leftMotorBackward(uint8_t speed, uint16_t distance = 0);
  //truck.leftMotorStop(bool hard = true);

  //Управление отдельно правым мотором
  //truck.rightMotorForward(uint8_t speed, uint16_t distance = 0);
  //truck.rightMotorBackward(uint8_t speed, uint16_t distance = 0);
  //truck.rightMotorStop(bool hard = true);

  //Управление платформой целиком
  //truck.moveForward(uint8_t speed, uint16_t distance = 0);
  //truck.moveBackward(uint8_t speed, uint16_t distance = 0);
  //truck.turnLeft(uint8_t speed, uint8_t radius, uint16_t angle);
  //truck.turnRight(uint8_t speed, uint8_t radius, uint16_t angle);
}
