#include "I2CServo.h"

//#define DEBUG_MODE_ALL
//#define DEBUG_MODE_ANGLE_AND_SPEED

I2CServo::I2CServo(
	PCA9685* pwmCtrlr,
	uint8_t channel,
	uint16_t min,
	uint16_t max,
	uint8_t minSpeed,
	uint16_t maxSpeed,
	uint8_t acceleration,
	uint8_t pwmFrequency
) {
	this->pwmController = pwmCtrlr;
	this->channel = channel;
	this->min = min;
	this->max = max;
	this->minAngularSpeed = minSpeed;
	this->maxAngularSpeed = maxSpeed;
	this->acceleration = acceleration;

	// LSB = (1s/freq)/4096 => 1 000 000us/(freq*4096) => 10 000/(freq*40.96)
	// Let's save RAM and use 41 instead 40.96. So LSB = 10 000/(freq*41)
	this->lsb = (float)10000/(pwmFrequency * 41);
}

void I2CServo::begin(uint8_t internalTimerMs) {
	this->internalTimerMs = internalTimerMs;

	if (internalTimerMs > 0) {
		Chronos::attachMsMemberCallback(this, internalTimerMs);
	}
}

// Degree/1 sec
void I2CServo::setMinAngularSpeed(uint8_t minSpeed) {
	this->minAngularSpeed = minSpeed;
}

// Degree/1 sec
void I2CServo::setMaxAngularSpeed(uint16_t maxSpeed) {
	this->maxAngularSpeed = maxSpeed;
}

// Degree/1 sec^2
void I2CServo::setMaxAcceleration(uint8_t acceleration) {
	this->acceleration = acceleration;
}


void I2CServo::smoothRotate(int16_t srcAngle, int16_t destAngle) {
	this->currentAngle = normalizeAngle(srcAngle);
	this->smoothRotate(destAngle);
}

void I2CServo::smoothRotate(int16_t angle) {
	angle = normalizeAngle(angle);

	if (this->currentAngle == -1 || this->currentAngle == angle) {
		#ifdef DEBUG_MODE_ALL
			Serial.println("I2CServo::smoothRotate: Aborded! Current angle is undefined");
		#endif

		return;
	}

	this->destAngle = angle;
	this->speedDownAngle = this->destAngle;
	this->middleAngle = (this->currentAngle + this->destAngle) / 2;
	this->currentAngularSpeedMsScale = (uint32_t)this->minAngularSpeed * MS_FACTOR;

	if (this->currentAngle > angle) {
		// CW rotation
		this->isCCW = false;
	} else {
		// CCW rotation
		this->isCCW = true;
	}
}

void I2CServo::rotate(int16_t angle) {
	angle = normalizeAngle(angle);
	this->destAngle = angle;
	setAngle(angle);
}

/**
 * Alias for rotate() function
 */
void I2CServo::write(int16_t angle) {
	this->rotate(angle);
}

void I2CServo::writeMicroseconds(uint16_t length) {
	length = min(this->max, max(this->min, length));
	length = (uint16_t)length/lsb;

	#ifdef DEBUG_MODE_ALL
		Serial.print("I2CServo::writeMicroseconds: LSB=");
		Serial.print(lsb); Serial.print(", Length="); Serial.println(length);
	#endif

	this->pwmController->setPWM(this->channel, 0, length);
}

bool I2CServo::isBusy() {
	return this->isRotationOver();
}

void I2CServo::chronosHandler() {
	this->externalTick(this->internalTimerMs);
}

void I2CServo::externalTick(uint8_t periodMs) {
	// STEP 1: Check is rotation still not over
	if (isRotationOver()) {
		// Rotation is already over
		return;
	}

	// STEP 2: calculate current angular speed
	calculateAngularSpeed(periodMs);

	// STEP 3: calculate new angle
	int16_t newAngle = getAngle(periodMs);

	if (newAngle != this->currentAngle) {
		setAngle(newAngle);

		#ifdef DEBUG_MODE_ALL
			Serial.print("New angle:"); Serial.println(newAngle);
		#endif

		#ifdef DEBUG_MODE_ANGLE_AND_SPEED
			Serial.print(newAngle);
			Serial.print("\t");
			Serial.println(this->currentAngularSpeedMsScale/1000);
		#endif
	}
}

/*************************************************************************/
/**************************    Private     *******************************/
/*************************************************************************/

void I2CServo::setAngle(int16_t angle) {
	this->currentAngle = angle;
	uint16_t length = this->min + (float)(this->max - this->min)/180 * angle;

	#ifdef DEBUG_MODE_ALL
		Serial.print("I2CServo::setAngle: Angle="); Serial.println(angle);
	#endif

	writeMicroseconds(length);
}

int16_t I2CServo::normalizeAngle(uint16_t angle) {
	return max(0, min(360, angle));
}

uint16_t I2CServo::calculateAngularSpeed(uint8_t timeMs) {
	uint32_t tmp32bit = 0;

	// Check is we still less than at middleAngle
	if (isMiddleAngleExceded() == false) {
		//Middle angle still not exceded.
		// Check if we still need speed up or not
		if (isMaxSpeedReached() == false) {
			tmp32bit = this->acceleration * timeMs;
			this->currentAngularSpeedMsScale = this->currentAngularSpeedMsScale + tmp32bit;
			
			tmp32bit = (uint32_t)this->minAngularSpeed * MS_FACTOR;
			this->currentAngularSpeedMsScale = max(tmp32bit, this->currentAngularSpeedMsScale);
		}

	} else {
		// Middle angle already exceded.
		// Check if we already need to slow speed down
		if (isSpeedDownAngleExceded()) {
			tmp32bit = this->acceleration * timeMs;
			this->currentAngularSpeedMsScale = this->currentAngularSpeedMsScale - min(this->currentAngularSpeedMsScale, tmp32bit);

			tmp32bit = (uint32_t)this->minAngularSpeed * MS_FACTOR;
			this->currentAngularSpeedMsScale = max(tmp32bit, this->currentAngularSpeedMsScale);
		}
	}

	#ifdef DEBUG_MODE_ALL
		Serial.print("Angular Speed:"); Serial.println(this->currentAngularSpeedMsScale/1000);
	#endif
}

int16_t I2CServo::getAngle(uint8_t timeMs) {
	this->angleChangeTime += timeMs;

	uint16_t delta = (uint16_t) ((this->currentAngularSpeedMsScale/MS_FACTOR) * this->angleChangeTime) / MS_FACTOR;

	if (delta > 0) {
		this->angleChangeTime = 0;

		if (isMaxSpeedReached() == false) {
			if (this->isCCW == true) {
				this->speedDownAngle -= delta;
			} else {
				this->speedDownAngle += delta;
			}
		}

		if (this->isCCW == true) {
			return min(this->destAngle, this->currentAngle + delta);
		} else {
			return max(this->destAngle, this->currentAngle - min(this->currentAngle, delta));
		}
	}

	return this->currentAngle;
}

bool I2CServo::isRotationOver() {
	if (this->currentAngle < 0)
		return false;

	if (this->isCCW == true) {
		return this->currentAngle >= this->destAngle;
	} else {
		return this->currentAngle <= this->destAngle;
	}
}

bool I2CServo::isMiddleAngleExceded() {
	if (this->middleAngle < 0 || this->currentAngle < 0)
		return false;

	if (this->isCCW == true) {
		return this->currentAngle >= this->middleAngle;
	} else {
		return this->currentAngle <= this->middleAngle;
	}
}

bool I2CServo::isSpeedDownAngleExceded() {
	if (this->speedDownAngle < 0 || this->currentAngle < 0)
		return false;

	if (this->isCCW == true) {
		return this->currentAngle >= this->speedDownAngle;
	} else {
		return this->currentAngle <= this->speedDownAngle;
	}
}

bool I2CServo::isMaxSpeedReached() {
	return this->currentAngularSpeedMsScale/MS_FACTOR >= this->maxAngularSpeed;
}
